@@ -1,15 +1,36 @@
+/* Copyright 2007, 2008 Peter Klausler.  See COPYING for license. */
 #include "all.h"
 #include <fcntl.h>
 #include <sys/mman.h>
 #include <sys/stat.h>
 
-int utf8_mode = 2; /* auto */
+enum utf8_mode utf8_mode = UTF8_AUTO;
 
-static int old_fashioned_read(struct text *text)
+const char *path_format(const char *path)
+{
+	char *cwdbuf, *cwd;
+	const char *slash;
+
+	if (!path || *path != '/')
+		return path;
+	cwdbuf = allocate(1024);
+	cwd = getcwd(cwdbuf, 1024);
+	while ((slash = strchr(path, '/')) &&
+	       !strncmp(cwd, path, slash - path)) {
+		cwd += slash++ - path;
+		if (*cwd && *cwd++ != '/')
+			break;
+		path = slash;
+	}
+	RELEASE(cwdbuf);
+	return path;
+}
+
+static ssize_t old_fashioned_read(struct text *text)
 {
 	char *raw;
-	int got, total = 0;
-	unsigned max;
+	ssize_t got, total = 0;
+	size_t max;
 #define CHUNK 1024
 
 	do {
@@ -18,7 +39,8 @@ static int old_fashioned_read(struct text *text)
 		errno = 0;
 		got = read(text->fd, raw, max);
 		if (got < 0) {
-			message("error reading %s", text->path);
+			message("%s: can't read",
+				path_format(text->path));
 			buffer_delete(text->buffer, 0, total + CHUNK);
 			return -1;
 		}
@@ -32,8 +54,8 @@ static int old_fashioned_read(struct text *text)
 static char *fix_path(const char *path)
 {
 	char *fpath;
-	int freepath = 0;
-	int pathlen;
+	const char *freepath = NULL, *home;
+	size_t pathlen;
 
 	if (!path)
 		return NULL;
@@ -49,33 +71,34 @@ static char *fix_path(const char *path)
 			;
 		if (!pathlen)
 			return NULL;
-		apath = allocate(NULL, pathlen+1);
+		apath = allocate(pathlen+1);
 		memcpy(apath, path, pathlen);
 		apath[pathlen] = '\0';
-		path = apath;
-		freepath = 1;
+		freepath = path = apath;
 	}
-	if (*path != '/') {
-		char *cwdbuf = allocate(NULL, 1024);
+	if (!strncmp(path, "~/", 2) && (home = getenv("HOME"))) {
+		char *apath = allocate(strlen(home) + pathlen);
+		sprintf(apath, "%s%s", home, path + 1);
+		RELEASE(freepath);
+		freepath = path = apath;
+	} else if (*path != '/') {
+		char *cwdbuf = allocate(1024);
 		char *cwd = getcwd(cwdbuf, 1024);
-		char *apath = allocate(NULL, strlen(cwd) + pathlen + 2);
+		char *apath = allocate(strlen(cwd) + pathlen + 2);
 		sprintf(apath, "%s/%s", cwd, path);
-		if (freepath)
-			allocate(path, 0);
-		allocate(cwdbuf, 0);
-		path = apath;
-		freepath = 1;
+		RELEASE(freepath);
+		RELEASE(cwdbuf);
+		freepath = path = apath;
 	}
 	fpath = strdup(path);
-	if (freepath)
-		allocate(path, 0);
+	RELEASE(freepath);
 	return fpath;
 }
 
-static void clean_mmap(struct text *text, unsigned bytes, int flags)
+static void clean_mmap(struct text *text, size_t bytes, int flags)
 {
 	void *p;
-	unsigned pagesize = getpagesize();
+	size_t pagesize = getpagesize();
 	unsigned pages = (bytes + pagesize - 1) / pagesize;
 
 	if (text->clean)
@@ -102,17 +125,20 @@ static void grab_mtime(struct text *text)
 static void scan(struct view *view)
 {
 	char *raw, scratch[8];
-	unsigned bytes = view_raw(view, &raw, 0, getpagesize());
-	unsigned at, chlen, ch, check, lastch = 0, crnl = 0, nl = 0;
-	unsigned chop = bytes < view->bytes ? 8 : 0;
+	position_t at;
+	size_t bytes = view_raw(view, &raw, 0, getpagesize());
+	size_t chop = bytes < view->bytes ? 8 : 0;
+	size_t chlen, check;
+	Unicode_t ch, lastch = 0;
+	int crnl = 0, nl = 0;
 
-	if (!utf8_mode)
+	if (utf8_mode == UTF8_NO)
 		view->text->flags |= TEXT_NO_UTF8;
-	else if (utf8_mode == 2 /*auto*/)
+	else if (utf8_mode == UTF8_AUTO)
 		for (at = 0; at + chop < bytes; at += chlen) {
 			chlen = utf8_length(raw + at, bytes - at);
 			ch = utf8_unicode(raw + at, chlen);
-			check = utf8_out(scratch, ch);
+			check = unicode_utf8(scratch, ch);
 			if (chlen != check) {
 				view->text->flags |= TEXT_NO_UTF8;
 				break;
@@ -153,20 +179,20 @@ struct view *view_open(const char *path0)
 	errno = 0;
 	if (stat(path, &statbuf)) {
 		if (errno != ENOENT) {
-			message("can't stat %s", path);
+			message("%s: can't stat", path_format(path));
 			goto fail;
 		}
 		errno = 0;
 		text->fd = open(path, O_CREAT|O_TRUNC|O_RDWR,
 				S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
 		if (text->fd < 0) {
-			message("can't create %s", path);
+			message("%s: can't create", path_format(path));
 			goto fail;
 		}
 		text->flags |= TEXT_CREATED;
 	} else {
 		if (!S_ISREG(statbuf.st_mode)) {
-			message("%s is not a regular file", path);
+			message("%s: not a regular file", path_format(path));
 			goto fail;
 		}
 		text->fd = open(path, O_RDWR);
@@ -175,7 +201,7 @@ struct view *view_open(const char *path0)
 			text->flags |= TEXT_RDONLY;
 			text->fd = open(path, O_RDONLY);
 			if (text->fd < 0) {
-				message("can't open %s", path);
+				message("%s: can't open", path_format(path));
 				goto fail;
 			}
 		}
@@ -196,11 +222,11 @@ struct view *view_open(const char *path0)
 fail:	view_close(view);
 	view = NULL;
 
-done:	allocate(path, 0);
+done:	RELEASE(path);
 	return view;
 }
 
-static int try_dir(char *path, const char *dir, const struct tm *gmt)
+static fd_t try_dir(char *path, const char *dir, const struct tm *gmt)
 {
 	struct stat statbuf;
 
@@ -227,7 +253,7 @@ struct view *text_new(void)
 	const char *me, *home;
 	time_t now = time(NULL);
 	struct tm *gmt = gmtime(&now);
-	int fd = -1;
+	fd_t fd = -1;
 	struct view *view;
 
 	if ((home = getenv("HOME"))) {
@@ -258,25 +284,25 @@ struct view *text_new(void)
 	return view;
 }
 
-int text_rename(struct text *text, const char *path0)
+Boolean_t text_rename(struct text *text, const char *path0)
 {
 	char *path = fix_path(path0);
 	struct text *b;
 	struct view *view;
-	int fd;
+	fd_t fd;
 
 	if (!path)
-		return 0;
+		return FALSE;
 	for (b = text; b; b = b->next)
 		if (b->path && !strcmp(b->path, path))
-			return 0;
+			return FALSE;
 
 	errno = 0;
 	if ((fd = open(path, O_CREAT|O_RDWR,
 			S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) < 0) {
-		message("can't create %s", path);
-		allocate(path, 0);
-		return 0;
+		message("%s: can't create", path_format(path));
+		RELEASE(path);
+		return FALSE;
 	}
 
 	if (text->flags & TEXT_CREATED) {
@@ -295,18 +321,19 @@ int text_rename(struct text *text, const char *path0)
 	}
 	text->fd = fd;
 	grab_mtime(text);
-	allocate(text->path, 0);
+	RELEASE(text->path);
 	text->path = path;
+	keyword_init(text);
 	for (view = text->views; view; view = view->next)
 		view_name(view);
-	return 1;
+	return TRUE;
 }
 
 void text_dirty(struct text *text)
 {
 	if (text->path && !text->dirties && text->flags & TEXT_RDONLY)
-		message("%s is a read-only file, changes will not be saved.",
-			text->path);
+		message("%s: read-only, changes won't be saved here",
+			path_format(text->path));
 	text->dirties++;
 	if (!text->buffer) {
 		text->buffer = buffer_create(text->fd >= 0 ? text->path : NULL);
@@ -319,31 +346,31 @@ void text_dirty(struct text *text)
 static void save_original(struct text *text)
 {
 	char *save_path;
-	int fd;
+	fd_t fd;
 
 	if (!text->clean ||
 	    !text->path ||
 	    text->flags & (TEXT_SAVED_ORIGINAL |
+			   TEXT_RDONLY |
 			   TEXT_CREATED |
 			   TEXT_EDITOR))
 		return;
 
-	save_path = allocate(NULL, strlen(text->path)+2);
+	save_path = allocate(strlen(text->path)+2);
 	sprintf(save_path, "%s~", text->path);
 	errno = 0;
 	fd = creat(save_path, S_IRUSR|S_IWUSR);
 	if (fd < 0)
-		message("can't save copy of original file to %s",
-			save_path);
+		message("%s: can't save original text", path_format(save_path));
 	else {
 		write(fd, text->clean, text->clean_bytes);
 		close(fd);
 	}
-	allocate(save_path, 0);
+	RELEASE(save_path);
 	text->flags |= TEXT_SAVED_ORIGINAL;
 }
 
-int text_is_dirty(struct text *text)
+Boolean_t text_is_dirty(struct text *text)
 {
 	return	text->preserved != text->dirties &&
 		text->fd >= 0 &&
@@ -353,17 +380,15 @@ int text_is_dirty(struct text *text)
 void text_preserve(struct text *text)
 {
 	char *raw;
-	unsigned bytes;
+	size_t bytes;
 	struct stat statbuf;
 
 	if (text->preserved == text->dirties || text->fd < 0 || !text->buffer)
 		return;
-
 	text->preserved = text->dirties;
-
 	text_unfold_all(text);
-
 	if (text->clean) {
+		save_original(text);
 		bytes = buffer_raw(text->buffer, &raw, 0, ~0);
 		if (bytes == text->clean_bytes &&
 		    !memcmp(text->clean, raw, bytes))
@@ -371,17 +396,36 @@ void text_preserve(struct text *text)
 		munmap(text->clean, text->clean_bytes);
 		text->clean = NULL;
 	}
-
-	save_original(text);
-
 	if (text->mtime &&
 	    text->path &&
 	    !fstat(text->fd, &statbuf) &&
 	    text->mtime < statbuf.st_mtime)
-		message("%s has been modified since it was read into the "
-			"editor, and those changes may have been overwritten.",
-			text->path);
-
+		message("%s: modified since read into the "
+			"editor, changes may have been overwritten.",
+			path_format(text->path));
+	text->preserved = text->dirties;
+	if (text->flags & TEXT_RDONLY && text->path) {
+		int newfd;
+		char *new_path = allocate(strlen(text->path) + 2);
+		sprintf(new_path, "%s@", text->path);
+		errno = 0;
+		newfd = open(new_path, O_CREAT|O_TRUNC|O_RDWR,
+			     S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
+		if (newfd < 0) {
+			message("%s: can't create", path_format(new_path));
+			RELEASE(new_path);
+			return;
+		}
+		message("%s: read-only, new version saved to %s@",
+			text->path, text->path);
+		text->flags &= ~TEXT_RDONLY;
+		text->flags |= TEXT_CREATED;
+		close(text->fd);
+		RELEASE(text->path);
+		text->fd = newfd;
+		text->path = new_path;
+	} else
+		text->flags &= ~TEXT_CREATED;
 	bytes = buffer_raw(text->buffer, &raw, 0, ~0);
 	ftruncate(text->fd, bytes);
 	clean_mmap(text, bytes, PROT_READ|PROT_WRITE);
@@ -389,12 +433,13 @@ void text_preserve(struct text *text)
 		memcpy(text->clean, raw, bytes);
 		msync(text->clean, bytes, MS_SYNC);
 	} else {
+		ssize_t wrote;
 		lseek(text->fd, 0, SEEK_SET);
-		write(text->fd, raw, bytes);
+		errno = 0;
+		wrote = write(text->fd, raw, bytes);
+		if (wrote != bytes)
+			message("%s: write failed", path_format(text->path));
 	}
-
-	text->preserved = text->dirties;
-	text->flags &= ~TEXT_CREATED;
 	grab_mtime(text);
 }
 
