@@ -1,28 +1,35 @@
+/* Copyright 2007, 2008 Peter Klausler.  See COPYING for license. */
 #include "all.h"
 
 /*
  *	This is the default command mode.
  */
 
-int is_asdfg;
+Boolean_t is_asdfg;
+
+static struct macro *default_macro, *function_key[FUNCTION_FKEYS+1];
 
 struct mode_default {
 	command command;
-	int variant, value, is_hex;
+	Boolean_t variant, is_hex;
+	int value;
 };
 
-static void command_handler(struct view *, unsigned);
+static void command_handler(struct view *, Unicode_t);
 
-static unsigned cut(struct view *view, int delete)
+
+static position_t cut(struct view *view, Boolean_t delete)
 {
 	struct mode_default *mode = (struct mode_default *) view->mode;
-	unsigned offset;
-	int append;
-	int bytes = view_get_selection(view, &offset, &append);
+	position_t offset;
+	Boolean_t append;
+	size_t bytes = view_get_selection(view, &offset, &append);
+	int copies = mode->value ? mode->value : 1;
 
-	if (!mode->variant)
+	if (!mode->variant || mode->value)
 		clip_init(0);
-	clip(0, view, offset, bytes, append);
+	while (copies--)
+		clip(0, view, offset, bytes, append);
 	if (delete)
 		view_delete(view, offset, bytes);
 	locus_set(view, MARK, UNSET);
@@ -32,7 +39,7 @@ static unsigned cut(struct view *view, int delete)
 static void paste(struct view *view)
 {
 	struct mode_default *mode = (struct mode_default *) view->mode;
-	unsigned cursor = locus_get(view, CURSOR);
+	position_t cursor = locus_get(view, CURSOR);
 	clip_paste(view, cursor, mode->value);
 	locus_set(view, MARK, /*old*/ cursor);
 }
@@ -41,12 +48,28 @@ static void forward_lines(struct view *view)
 {
 	struct mode_default *mode = (struct mode_default *) view->mode;
 	unsigned count = mode->variant ? mode->value : 1;
-	unsigned cursor = locus_get(view, CURSOR);
+	position_t cursor = locus_get(view, CURSOR);
+
 	if (!count)
 		cursor = find_paragraph_end(view, cursor);
 	else
-		while (count-- && cursor < view->bytes)
+		for (; count && cursor < view->bytes; count--)
 			cursor = find_line_end(view, cursor+1);
+	if (count)
+		macros_abort();
+	locus_set(view, CURSOR, cursor);
+}
+
+static void down_lines(struct view *view)
+{
+	struct mode_default *mode = (struct mode_default *) view->mode;
+	unsigned count = mode->variant ? mode->value : 1;
+	position_t cursor = locus_get(view, CURSOR);
+
+	for (; count && cursor < view->bytes; count--)
+		cursor = find_line_down(view, cursor);
+	if (count)
+		macros_abort();
 	locus_set(view, CURSOR, cursor);
 }
 
@@ -54,12 +77,28 @@ static void backward_lines(struct view *view)
 {
 	struct mode_default *mode = (struct mode_default *) view->mode;
 	unsigned count = mode->variant ? mode->value : 1;
-	unsigned cursor = locus_get(view, CURSOR);
+	position_t cursor = locus_get(view, CURSOR);
+
 	if (!count)
 		cursor = find_paragraph_start(view, cursor);
 	else
-		while (count-- && cursor)
+		for (; count && cursor; count--)
 			cursor = find_line_start(view, cursor-1);
+	if (count)
+		macros_abort();
+	locus_set(view, CURSOR, cursor);
+}
+
+static void up_lines(struct view *view)
+{
+	struct mode_default *mode = (struct mode_default *) view->mode;
+	unsigned count = mode->variant ? mode->value : 1;
+	position_t cursor = locus_get(view, CURSOR);
+
+	for (; count && cursor; count--)
+		cursor = find_line_up(view, cursor);
+	if (count && !cursor)
+		macros_abort();
 	locus_set(view, CURSOR, cursor);
 }
 
@@ -67,9 +106,12 @@ static void forward_chars(struct view *view)
 {
 	struct mode_default *mode = (struct mode_default *) view->mode;
 	unsigned count = mode->variant ? mode->value : 1;
-	unsigned cursor = locus_get(view, CURSOR), next;
-	while (count-- && view_char(view, cursor, &next) >= 0)
+	position_t cursor = locus_get(view, CURSOR), next;
+
+	for (; count && IS_UNICODE(view_char(view, cursor, &next)); count--)
 		cursor = next;
+	if (count)
+		macros_abort();
 	locus_set(view, CURSOR, cursor);
 }
 
@@ -77,41 +119,74 @@ static void backward_chars(struct view *view)
 {
 	struct mode_default *mode = (struct mode_default *) view->mode;
 	unsigned count = mode->variant ? mode->value : 1;
-	unsigned cursor = locus_get(view, CURSOR);
-	while (count-- && cursor)
+	position_t cursor = locus_get(view, CURSOR);
+
+	for (; count && cursor; count--)
 		view_char_prior(view, cursor, &cursor);
+	if (count)
+		macros_abort();
 	locus_set(view, CURSOR, cursor);
 }
 
-static int funckey(struct view *view, int Fk)
+static Boolean_t funckey(struct view *view, int Fk)
 {
 	struct mode_default *mode = (struct mode_default *) view->mode;
-	if (Fk > FUNCTION_KEYS)
-		return 0;
+
+	if (Fk > FUNCTION_FKEYS)
+		return FALSE;
 	if (mode->variant && !mode->value) {
 		macro_end_recording(CONTROL('@'));
 		macro_free(function_key[Fk]);
 		function_key[Fk] = macro_record();
-		return 1;
+		return TRUE;
 	}
 	return macro_play(function_key[Fk], mode->value);
 }
 
-static void command_handler(struct view *view, unsigned ch0)
+static position_t self_insert(struct view *view, Unicode_t ch,
+			      position_t mark, position_t old_cursor,
+			      Boolean_t literal_unicode)
+{
+	char cbuf[8];
+	position_t cursor = old_cursor;
+	if (mark != UNSET && mark > cursor) {
+		cursor = cut(view, 1);
+		mark = UNSET;
+	}
+	if (ch == '\n' && view->text->flags & TEXT_CRNL) {
+		view_insert(view, "\r\n", cursor, 2);
+		cursor += 2;
+	} else if (ch <= 0x100 &&
+		   (!literal_unicode ||
+		    view->text->flags & TEXT_NO_UTF8)) {
+		cbuf[0] = ch;
+		view_insert(view, cbuf, cursor++, 1);
+	} else {
+		size_t len = unicode_utf8(cbuf, ch);
+		view_insert(view, cbuf, cursor, len);
+		cursor += len;
+	}
+	if (mark == old_cursor)
+		locus_set(view, MARK, old_cursor);
+	if (view->shell_std_in >= 0)
+		shell_command(view, ch);
+	return cursor;
+}
+
+static void command_handler(struct view *view, Unicode_t ch0)
 {
 	struct mode_default *mode = (struct mode_default *) view->mode;
-	unsigned ch = ch0;
-	unsigned cursor = locus_get(view, CURSOR);
-	unsigned mark = locus_get(view, MARK);
-	unsigned offset;
-	char buf[8];
-	int ok = 1, literal_unicode = 0;
+	Unicode_t ch = ch0;
+	position_t cursor = locus_get(view, CURSOR);
+	position_t mark = locus_get(view, MARK);
+	position_t offset;
+	Boolean_t ok = TRUE;
 	struct view *new_view;
 	char *select;
 
 	/* Backspace always deletes the character before cursor. */
 	if (ch == 0x7f /*BCK*/) {
-delete:		if (view_char_prior(view, cursor, &mark) >= 0)
+delete:		if (IS_UNICODE(view_char_prior(view, cursor, &mark)))
 			view_delete(view, mark, cursor-mark);
 		else
 			window_beep(view);
@@ -119,41 +194,46 @@ delete:		if (view_char_prior(view, cursor, &mark) >= 0)
 	}
 
 	/* Decode function-key sequences */
-	if (DISPLAY_IS_FKEY(ch)) {
+	if (IS_FUNCTION_KEY(ch)) {
+
+		/* Forget the up/down goal column if not moving up/down */
+		if (ch != FUNCTION_UP && ch != FUNCTION_DOWN)
+			view->goal.cursor = UNSET;
+
 		switch (ch) {
-		case DISPLAY_UP:
-			backward_lines(view);
+		case FUNCTION_DOWN:
+			down_lines(view);
 			break;
-		case DISPLAY_DOWN:
-			forward_lines(view);
+		case FUNCTION_UP:
+			up_lines(view);
 			break;
-		case DISPLAY_LEFT:
+		case FUNCTION_LEFT:
 			backward_chars(view);
 			break;
-		case DISPLAY_RIGHT:
+		case FUNCTION_RIGHT:
 			forward_chars(view);
 			break;
-		case DISPLAY_PGUP:
+		case FUNCTION_PGUP:
 			window_page_up(view);
 			break;
-		case DISPLAY_PGDOWN:
+		case FUNCTION_PGDOWN:
 			window_page_down(view);
 			break;
-		case DISPLAY_HOME:
+		case FUNCTION_HOME:
 			locus_set(view, CURSOR, 0);
 			break;
-		case DISPLAY_END:
+		case FUNCTION_END:
 			locus_set(view, CURSOR, view->bytes);
 			break;
-		case DISPLAY_INSERT:
+		case FUNCTION_INSERT:
 			paste(view);
 			break;
-		case DISPLAY_DELETE:
+		case FUNCTION_DELETE:
 			goto delete;
 		default:
-			if (ch < DISPLAY_F1 ||
-			    !funckey(view, ch - DISPLAY_F1 + 1))
-				ok = 0;
+			if (ch < FUNCTION_F(1) ||
+			    !funckey(view, ch - FUNCTION_F(1) + 1))
+				ok = FALSE;
 		}
 		goto done;
 	}
@@ -169,6 +249,8 @@ delete:		if (view_char_prior(view, cursor, &mark) >= 0)
 
 	if (ch >= ' ' /*0x20*/) {
 
+		view->goal.cursor = UNSET;
+
 		if (mode->variant) {
 			if (mode->is_hex && isxdigit(ch)) {
 				mode->value *= 16;
@@ -199,7 +281,7 @@ delete:		if (view_char_prior(view, cursor, &mark) >= 0)
 						locus_set(new_view, MARK, mark);
 					window_activate(new_view);
 				} else
-					ok = 0;
+					ok = FALSE;
 				goto done;
 			case ';':
 				window_after(view, text_new(), -1);
@@ -223,7 +305,7 @@ delete:		if (view_char_prior(view, cursor, &mark) >= 0)
 				else {
 					mark = view_unfold(view, cursor);
 					if ((signed) mark < 0)
-						window_beep(view);
+						view_unfold_all(view);
 					else
 						locus_set(view, MARK, mark);
 				}
@@ -231,24 +313,7 @@ delete:		if (view_char_prior(view, cursor, &mark) >= 0)
 			}
 		}
 
-		/* self insertion or replacement */
-self_insert:	if (mark != UNSET && mark > cursor) {
-			cursor = cut(view, 1);
-			mark = UNSET;
-		}
-		if (ch == '\n' && view->text->flags & TEXT_CRNL)
-			view_insert(view, "\r\n", cursor, 2);
-		else if (ch <= 0x100 &&
-			 (!literal_unicode ||
-			  view->text->flags & TEXT_NO_UTF8)) {
-			buf[0] = ch;
-			view_insert(view, buf, cursor, 1);
-		} else
-			view_insert(view, buf, cursor, utf8_out(buf, ch));
-		if (mark == cursor)
-			locus_set(view, MARK, /*old*/ cursor);
-		else if (ch == '\n')
-			shell_command(view);
+		self_insert(view, ch, mark, cursor, FALSE);
 		goto done;
 	}
 
@@ -268,6 +333,11 @@ self_insert:	if (mark != UNSET && mark > cursor) {
 		ch = asdfg_to_aoeui[ch-'A'];
 	}
 
+	if (ch != 'G' && ch != 'C' &&
+	    (ch != 'H' && ch != 'T' ||
+	     !mode->variant || mode->value))
+		view->goal.cursor = UNSET;
+
 	switch (ch) {
 	case '@': /* (^Space) */
 		if (mode->variant)
@@ -280,9 +350,9 @@ self_insert:	if (mark != UNSET && mark > cursor) {
 		break;
 	case 'B': /* exchange clip buffer and selection, if any, else paste */
 		if (mark != UNSET) {
-			int outbytes = view_get_selection(view, &offset, NULL);
+			size_t outbytes = view_get_selection(view, &offset, NULL);
 			unsigned reg = mode->value;
-			int inbytes = clip_paste(view, offset + outbytes, reg);
+			size_t inbytes = clip_paste(view, offset + outbytes, reg);
 			clip_init(reg);
 			clip(reg, view, offset, outbytes, 0);
 			view_delete(view, offset, outbytes);
@@ -297,42 +367,43 @@ self_insert:	if (mark != UNSET && mark > cursor) {
 	case 'D': /* [select whitespace] / cut [pre/appending] */
 		if (mark == UNSET && mode->variant) {
 			locus_set(view, MARK, find_nonspace(view, cursor));
-			while ((ch = view_char_prior(view, cursor, &offset))
-					>= 0 &&
-			       isspace(ch))
-				cursor = offset;
-			locus_set(view, CURSOR, cursor);
+			locus_set(view, CURSOR,
+				  find_nonspace_prior(view, cursor));
 		} else
-			cut(view, 1);
+			cut(view, TRUE);
 		break;
 	case 'E':
-		if (mark == UNSET)
-			if (mode->variant)
-				demultiplex_view(view);
-			else if (view->shell_std_in < 0) {
-				new_view = text_new();
-				window_after(view, new_view, -1);
-				mode_shell_pipe(new_view);
-			} else
-				window_beep(view);
-		else
+		if (mark != UNSET)
 			mode_child(view);
+		else if (mode->variant)
+			demultiplex_view(view);
+		else {
+			new_view = text_new();
+			window_after(view, new_view, -1);
+			mode_shell_pipe(new_view);
+		}
 		break;
 	case 'F': /* copy [pre/appending] */
-		cut(view, 0);
+		cut(view, FALSE);
 		break;
 	case 'G':
 		backward_lines(view);
 		break;
 	case 'H':
-		backward_chars(view);
+		if (mode->variant && !mode->value) {
+			mode->variant = FALSE;
+			up_lines(view);
+		} else
+			backward_chars(view);
 		break;
 	case 'I': /* (TAB) tab / tab completion [align; set tab stop] */
 		if (!mode->variant) {
 			if (!tab_completion_command(view))
 				insert_tab(view);
 		} else if (mode->value)
-			if (mode->value >= 1 && mode->value <= 20)
+			if (mode->value == 1)
+				view->text->flags ^= TEXT_NO_TABS;
+			else if (mode->value > 1 && mode->value <= 20)
 				view->text->tabstop =
 					default_tab_stop = mode->value;
 			else
@@ -340,11 +411,11 @@ self_insert:	if (mark != UNSET && mark > cursor) {
 		else
 			align(view);
 		break;
-	case 'J': /* line feed: new line with alignment */
-		if (view->text->flags & TEXT_CRNL)
-			view_insert(view, "\r", cursor++, 1);
-		view_insert(view, "\n", cursor++, 1);
-		align(view);
+	case 'J':  /* line feed: new line */
+	case 'M': /* (ENTER) new line with alignment */
+		self_insert(view, '\n', mark, cursor, FALSE);
+		if (ch == 'J')
+			align(view);
 		break;
 	case 'K': /* save all [single] */
 		if (mode->variant)
@@ -358,21 +429,9 @@ self_insert:	if (mark != UNSET && mark > cursor) {
 		else
 			window_page_down(view);
 		break;
-	case 'M': /* (ENT) new line [opened] */
-		if (mode->variant) {
-			if (view->text->flags & TEXT_CRNL)
-				view_insert(view, "\r\n", cursor, 2);
-			else
-				view_insert(view, "\n", cursor, 1);
-			locus_set(view, CURSOR, cursor);
-		} else {
-			ch = '\n';
-			goto self_insert;
-		}
-		break;
 	case 'N': /* backward word(s) [sentence] */
 		if (mode->value)
-			while (mode->value--)
+			while (mode->value-- && cursor)
 				cursor = find_word_start(view, cursor);
 		else if (mode->variant)
 			cursor = find_sentence_start(view, cursor);
@@ -383,11 +442,10 @@ self_insert:	if (mark != UNSET && mark > cursor) {
 	case 'O': /* macro end/execute [start] */
 		if (mode->variant && !mode->value) {
 			macro_end_recording(CONTROL('@'));
-			macro_free(view->local_macro);
-			view->local_macro = macro_record();
+			macro_free(default_macro);
+			default_macro = macro_record();
 		} else if (!macro_end_recording(ch0) &&
-			   !macro_play(view->local_macro,
-				       mode->value))
+			   !macro_play(default_macro, mode->value))
 			window_beep(view);
 		break;
 	case 'P': /* select other window [closing current] */
@@ -426,11 +484,14 @@ self_insert:	if (mark != UNSET && mark > cursor) {
 		locus_set(view, CURSOR, cursor);
 		break;
 	case 'T':
-		forward_chars(view);
+		if (mode->variant && !mode->value) {
+			mode->variant = FALSE;
+			down_lines(view);
+		} else
+			forward_chars(view);
 		break;
 	case 'U': /* undo [redo] */
-		offset = (mode->variant ? text_redo :
-					  text_undo)(view->text);
+		offset = (mode->variant ? text_redo : text_undo)(view->text);
 		if ((offset -= view->start) <= view->bytes)
 			locus_set(view, CURSOR, offset);
 		locus_set(view, MARK, UNSET);
@@ -465,7 +526,7 @@ self_insert:	if (mark != UNSET && mark > cursor) {
 				new_view = NULL;
 			} else
 				ok = !!(new_view = view_open(select));
-			allocate(select, 0);
+			RELEASE(select);
 			if (ok)
 				view_delete_selection(view);
 			if (new_view)
@@ -474,9 +535,9 @@ self_insert:	if (mark != UNSET && mark > cursor) {
 		break;
 	case 'Y': /* split window [vertically] */
 		if (mark != UNSET) {
-			unsigned offset = mark < cursor ? mark : cursor;
-			unsigned bytes = mark < cursor ? cursor - mark :
-					 mark - cursor;
+			position_t offset = mark < cursor ? mark : cursor;
+			size_t bytes = mark < cursor ? cursor - mark :
+				       mark - cursor;
 			new_view = view_selection(view, offset, bytes);
 		} else
 			new_view = view_next(view);
@@ -508,28 +569,26 @@ self_insert:	if (mark != UNSET && mark > cursor) {
 			locus_set(view, CURSOR, cursor);
 		break;
 	case '^': /* literal [; unicode] */
-		if (mode->value) {
-			ch = mode->value;
-			literal_unicode = 1;
-			goto self_insert;
-		}
-		if ((signed) (ch = macro_getch()) >= 0) {
+		if (mode->value)
+			self_insert(view, mode->value, mark, cursor, TRUE);
+		else if (IS_UNICODE(ch = macro_getch())) {
 			if (ch >= '@' && ch <= '_')
 				ch = CONTROL(ch);
 			else if (ch >= 'a' && ch <= 'z')
 				ch = CONTROL(ch-'a'+'A');
 			else if (ch == '?')
 				ch = 0x7f;
-			goto self_insert;
-		}
-		ok = 0;
+			self_insert(view, ch, mark, cursor, FALSE);
+		} else
+			ok = FALSE;
 		break;
 	default:
-		ok = 0;
+		ok = FALSE;
 		break;
 	}
 
-done:	mode->variant = mode->value = mode->is_hex = 0;
+done:	mode->variant = mode->is_hex = FALSE;
+	mode->value = 0;
 	if (!ok)
 		window_beep(view);
 }
