@@ -1,3 +1,4 @@
+/* Copyright 2007, 2008 Peter Klausler.  See COPYING for license. */
 #include "all.h"
 #include <regex.h>
 
@@ -8,18 +9,18 @@
 struct mode_search {
 	command command;
 	struct mode *previous;
-	unsigned char *pattern;
-	unsigned bytes, alloc, last_bytes;
-	int backward;
-	unsigned start, mark;
+	Byte_t *pattern;
+	size_t bytes, alloc, last_bytes;
+	Boolean_t backward;
+	position_t start, mark;
 	regex_t *regex;
-	int regex_ready;
+	Boolean_t regex_ready;
 };
 
-static int match_char(int x, int y)
+static Boolean_t match_char(Unicode_t x, Unicode_t y)
 {
-	if (x < 0 || y < 0)
-		return 0;
+	if (!IS_UNICODE(x) || !IS_UNICODE(y))
+		return FALSE;
 	if (x >= 'a' && x <= 'z')
 		x += 'A' - 'a';
 	if (y >= 'a' && y <= 'z')
@@ -27,10 +28,11 @@ static int match_char(int x, int y)
 	return x == y;
 }
 
-static int match_pattern(struct view *view, unsigned offset)
+static size_t match_pattern(struct view *view, position_t offset)
 {
 	struct mode_search *mode = (struct mode_search *) view->mode;
-	unsigned j;
+	size_t j;
+
 	for (j = 0; j < mode->bytes; j++)
 		if (!match_char(mode->pattern[j],
 				view_byte(view, offset++)))
@@ -38,11 +40,11 @@ static int match_pattern(struct view *view, unsigned offset)
 	return mode->bytes;
 }
 
-static int match_regex(struct view *view, unsigned *offset, int advance)
+static int match_regex(struct view *view, position_t*offset, Boolean_t advance)
 {
 	int j, err;
 	char *raw;
-	unsigned bytes = view_raw(view, &raw, *offset, ~0);
+	size_t bytes = view_raw(view, &raw, *offset, ~0);
 	unsigned flags = 0;
 	regmatch_t match[10];
 	struct mode_search *mode = (struct mode_search *) view->mode;
@@ -74,8 +76,8 @@ static int match_regex(struct view *view, unsigned *offset, int advance)
 	return match[0].rm_eo - match[0].rm_so;
 }
 
-static int scan_forward(struct view *view, unsigned *length,
-			unsigned offset, unsigned max_offset)
+static int scan_forward(struct view *view, size_t *length,
+			position_t offset, position_t max_offset)
 {
 	struct mode_search *mode = (struct mode_search *) view->mode;
 
@@ -94,8 +96,8 @@ static int scan_forward(struct view *view, unsigned *length,
 	return -1;
 }
 
-static int scan_backward(struct view *view, unsigned *length,
-			 unsigned offset, unsigned min_offset)
+static int scan_backward(struct view *view, size_t *length,
+			 position_t offset, position_t min_offset)
 {
 	struct mode_search *mode = (struct mode_search *) view->mode;
 
@@ -114,31 +116,33 @@ static int scan_backward(struct view *view, unsigned *length,
 	return -1;
 }
 
-static int search(struct view *view, int backward, int new)
+static Boolean_t search(struct view *view, int backward, int new)
 {
 	struct mode_search *mode = (struct mode_search *) view->mode;
-	unsigned cursor, length;
+	position_t cursor;
+	size_t length;
 	int at;
 
 	if (!mode->bytes) {
 		locus_set(view, CURSOR, mode->start);
 		locus_set(view, MARK, UNSET);
-		return 1;
+		return TRUE;
 	}
 
 	if (mode->regex) {
 		int err;
 		if (new && mode->regex_ready) {
 			regfree(mode->regex);
-			mode->regex_ready = 0;
+			mode->regex_ready = FALSE;
 		}
 		if (!mode->regex_ready) {
 			mode->pattern[mode->bytes] = '\0';
+			status("regular expression: %s", mode->pattern);
 			err = regcomp(mode->regex, (char *) mode->pattern,
 				      REG_EXTENDED | REG_ICASE | REG_NEWLINE);
 			if (err)
-				return 0;
-			mode->regex_ready = 1;
+				return FALSE;
+			mode->regex_ready = TRUE;
 		}
 	}
 
@@ -165,10 +169,10 @@ static int search(struct view *view, int backward, int new)
 	locus_set(view, CURSOR, at);
 	locus_set(view, MARK, at + length);
 	mode->last_bytes = mode->bytes;
-	return 1;
+	return TRUE;
 }
 
-static void command_handler(struct view *view, unsigned ch)
+static void command_handler(struct view *view, Unicode_t ch)
 {
 	struct mode_search *mode = (struct mode_search *) view->mode;
 	static char cmdchar[][2] = {
@@ -177,6 +181,8 @@ static void command_handler(struct view *view, unsigned ch)
 		{ CONTROL('V'), CONTROL('U') }
 	};
 
+	static char *last_search;
+
 	/* Backspace removes the last character from the search target and
 	 * returns to the previous hit
 	 */
@@ -195,13 +201,13 @@ static void command_handler(struct view *view, unsigned ch)
 	 * match the extended target.
 	 */
 	if (ch >= ' ' || ch == '\t') {
-		unsigned new;
+		size_t new;
 		if (mode->bytes + 8 > mode->alloc) {
 			mode->alloc = mode->bytes + 64;
-			mode->pattern = allocate(mode->pattern, mode->alloc);
+			mode->pattern = reallocate(mode->pattern, mode->alloc);
 		}
-		mode->bytes += new = utf8_out((char *) mode->pattern +
-					      mode->bytes, ch);
+		mode->bytes += new = unicode_utf8((char *) mode->pattern +
+						  mode->bytes, ch);
 		mode->pattern[mode->bytes] = '\0';
 		if (!search(view, mode->backward, new) &&
 		    !mode->regex)
@@ -209,7 +215,7 @@ static void command_handler(struct view *view, unsigned ch)
 		return;
 	}
 
-	/* ^H moves to hit that is earlier in the text.
+	/* ^H moves to a hit that is earlier in the text.
 	 * ^T and ^/ (^A, ^_) proceed to a later hit.
 	 */
 	if (mode->last_bytes &&
@@ -218,50 +224,62 @@ static void command_handler(struct view *view, unsigned ch)
 	     ch == CONTROL('A') ||
 	     ch == CONTROL('_'))) {
 		mode->bytes = mode->last_bytes;
-		search(view, mode->backward = (ch == cmdchar[is_asdfg][0]), 0);
+		search(view, mode->backward = (ch == cmdchar[0][is_asdfg]), 0);
 		return;
 	}
 
-	/* Hitting ^T or ^/ (^A, ^_) with an empty search pattern causes
+	/* Hitting ^H/^T or ^/ (^A, ^_) with an empty search pattern causes
 	 * the last successful search target to be reused.
 	 */
-	if ((ch == cmdchar[1][is_asdfg] ||
+	if ((ch == cmdchar[0][is_asdfg] ||
+	     ch == cmdchar[1][is_asdfg] ||
 	     ch == CONTROL('A') ||
 	     ch == CONTROL('_')) &&
-	    !mode->bytes && view->last_search) {
-		mode->bytes = strlen(view->last_search);
+	    !mode->bytes && last_search) {
+		mode->bytes = strlen(last_search);
 		mode->alloc = mode->bytes + 8;
-		mode->pattern = allocate(mode->pattern, mode->alloc);
-		memcpy(mode->pattern, view->last_search, mode->bytes+1);
+		mode->pattern = reallocate(mode->pattern, mode->alloc);
+		memcpy(mode->pattern, last_search, mode->bytes+1);
+		if (ch == cmdchar[0][is_asdfg])
+			mode->backward = 1;
+		else if (ch == cmdchar[1][is_asdfg])
+			mode->backward = 0;
 		search(view, mode->backward, 0);
 		return;
 	}
 
 	/* Search is done */
 done:	if (mode->bytes) {
-		view->last_search = allocate(view->last_search, mode->bytes+1);
-		memcpy(view->last_search, mode->pattern, mode->bytes);
-		view->last_search[mode->bytes] = '\0';
+		last_search = reallocate(last_search, mode->bytes+1);
+		memcpy(last_search, mode->pattern, mode->bytes);
+		last_search[mode->bytes] = '\0';
 	}
 
 	view->mode = mode->previous;
+	status_hide();
 
 	if (ch == '\r' || ch == CONTROL('A') || ch == CONTROL('_'))
 		;
-	else if (ch == cmdchar[2][is_asdfg]) /* unset mark */
+	else if (ch == cmdchar[2][is_asdfg]) {
+		/* restore mark */
+		if (mode->mark != UNSET && !mode->backward) {
+			position_t endmark = locus_get(view, MARK);
+			if (endmark != UNSET)
+				locus_set(view, CURSOR, endmark);
+		}
 		locus_set(view, MARK, mode->mark);
-	else if (ch != 0x7f /*BCK*/)
+	} else if (ch != 0x7f /*BCK*/)
 		view->mode->command(view, ch);
 
 	/* Release search mode resources */
-	allocate(mode->pattern, 0);
+	RELEASE(mode->pattern);
 	if (mode->regex_ready)
 		regfree(mode->regex);
-	allocate(mode->regex, 0);
-	allocate(mode, 0);
+	RELEASE(mode->regex);
+	RELEASE(mode);
 }
 
-void mode_search(struct view *view, int regex)
+void mode_search(struct view *view, Boolean_t regex)
 {
 	struct mode_search *mode = allocate0(sizeof *mode);
 	mode->previous = view->mode;
