@@ -1,3 +1,4 @@
+/* Copyright 2007, 2008 Peter Klausler.  See COPYING for license. */
 #include "all.h"
 #include <sys/mman.h>
 
@@ -11,8 +12,8 @@
  */
 
 struct text *text_list;
-int default_tab_stop = 8; /* dammit */
-int no_tabs;
+unsigned default_tab_stop = 8; /* the only correct value :-) */
+Boolean_t default_no_tabs;
 
 struct view *view_find(const char *name)
 {
@@ -37,7 +38,7 @@ void view_name(struct view *view)
 	if (!name)
 		name = "";
 	len = strlen(name);
-	new = allocate(NULL, len + 8);
+	new = allocate(len + 8);
 	if (view->text->flags & TEXT_EDITOR) {
 		memcpy(new, name, len+1);
 		p = new + len;
@@ -64,7 +65,7 @@ void view_name(struct view *view)
 				break;
 		}
 
-	allocate(view->name, 0);
+	RELEASE(view->name);
 	view->name = new;
 }
 
@@ -72,7 +73,7 @@ struct view *view_create(struct text *text)
 {
 	struct view *view = allocate0(sizeof *view);
 	view->loci = DEFAULT_LOCI;
-	view->locus = allocate(NULL, view->loci * sizeof *view->locus);
+	view->locus = allocate(view->loci * sizeof *view->locus);
 	memset(view->locus, UNSET, view->loci * sizeof *view->locus);
 	view->locus[CURSOR] = 0;
 	view->text = text;
@@ -81,7 +82,9 @@ struct view *view_create(struct text *text)
 	view->bytes = text->buffer ? buffer_bytes(text->buffer) :
 			text->clean ? text->clean_bytes :0;
 	view->mode = mode_default();
-	view->shell_std_in = view->shell_out_locus = -1;
+	view->shell_std_in = -1;
+	view->shell_pg = -1;
+	view->shell_out_locus = NO_LOCUS;
 	view_name(view);
 	return view;
 }
@@ -89,10 +92,15 @@ struct view *view_create(struct text *text)
 struct view *text_create(const char *path, unsigned flags)
 {
 	struct text *text = allocate0(sizeof *text), *prev, *bp;
+
+	if (default_no_tabs)
+		flags |= TEXT_NO_TABS;
 	text->tabstop = default_tab_stop;
 	text->fd = -1;
 	text->flags = flags;
 	text->path = strdup(path);
+	keyword_init(text);
+
 	for (prev = NULL, bp = text_list; bp; prev = bp, bp = bp->next)
 		;
 	if (prev)
@@ -123,18 +131,18 @@ static void text_close(struct text *text)
 		close(text->fd);
 	if (text->flags & (TEXT_SCRATCH | TEXT_CREATED))
 		unlink(text->path);
-	allocate(text->path, 0);
-	allocate(text, 0);
+	RELEASE(text->path);
+	RELEASE(text);
 }
 
 void view_close(struct view *view)
 {
 	struct view *vp, *prev = NULL;
-	struct text *text = view->text;
+	struct text *text;
 
 	if (!view)
 		return;
-
+	text = view->text;
 	window_destroy(view->window);
 	demultiplex_view(view);
 	bookmark_unset_view(view);
@@ -149,13 +157,12 @@ void view_close(struct view *view)
 				break;
 			}
 
-	allocate(view->name, 0);
-	allocate(view->last_search, 0);
-	macro_free(view->local_macro);
-	allocate(view, 0);
+	RELEASE(view->name);
+	RELEASE(view);
 }
 
-struct view *view_selection(struct view *current, unsigned offset, unsigned bytes)
+struct view *view_selection(struct view *current, position_t offset,
+			    size_t bytes)
 {
 	struct view *view;
 
@@ -173,7 +180,7 @@ struct view *view_selection(struct view *current, unsigned offset, unsigned byte
 	return view;
 }
 
-void text_adjust_loci(struct text *text, unsigned offset, int delta)
+void text_adjust_loci(struct text *text, position_t offset, int delta)
 {
 	struct view *view;
 
@@ -181,19 +188,19 @@ void text_adjust_loci(struct text *text, unsigned offset, int delta)
 		return;
 
 	if (delta < 0) {
-		unsigned limit = offset - delta;
+		position_t limit = offset - delta;
 		for (view = text->views; view; view = view->next)
 			if (limit < view->start)
 				view->start += delta;
 			else if (offset < view->start) {
-				unsigned loss = limit - view->start;
+				size_t loss = limit - view->start;
 				if (loss > view->bytes)
 					loss = view->bytes;
 				view->start = offset;
 				view->bytes -= loss;
 				loci_adjust(view, 0, -loss);
 			} else if (offset < view->start + view->bytes) {
-				unsigned loss = view->start + view->bytes -
+				size_t loss = view->start + view->bytes -
 						offset;
 				if (loss > -delta)
 					loss = -delta;
@@ -210,10 +217,9 @@ void text_adjust_loci(struct text *text, unsigned offset, int delta)
 			}
 }
 
-unsigned view_get(struct view *view, void *out, unsigned offset, unsigned bytes)
+static size_t text_get(struct text *text, void *out, position_t offset,
+		       size_t bytes)
 {
-	struct text *text = view->text;
-	offset += view->start;
 	if (text->buffer)
 		return buffer_get(text->buffer, out, offset, bytes);
 	if (!text->clean || offset >= text->clean_bytes)
@@ -224,28 +230,48 @@ unsigned view_get(struct view *view, void *out, unsigned offset, unsigned bytes)
 	return bytes;
 }
 
-unsigned view_raw(struct view *view, char **out, unsigned offset,
-		  unsigned bytes)
+size_t view_get(struct view *view, void *out, position_t offset, size_t bytes)
+{
+	if (offset >= view->bytes)
+		return 0;
+	if (offset + bytes > view->bytes)
+		bytes = view->bytes - offset;
+	return text_get(view->text, out, view->start + offset, bytes);
+}
+
+static size_t text_raw(struct text *text, char **out, position_t offset,
+		       size_t bytes)
 {
-	struct text *text = view->text;
-	offset += view->start;
 	if (text->buffer)
 		return buffer_raw(text->buffer, out, offset, bytes);
-	if (!text->clean || offset >= text->clean_bytes)
+	if (!text->clean) {
+		*out = NULL;
 		return 0;
+	}
 	if (offset + bytes > text->clean_bytes)
 		bytes = text->clean_bytes - offset;
-	*out = text->clean + offset;
+	*out = bytes ? text->clean + offset : NULL;
 	return bytes;
 }
 
-unsigned view_delete(struct view *view, unsigned offset, unsigned bytes)
+size_t view_raw(struct view *view, char **out, position_t offset, size_t bytes)
+{
+	if (offset >= view->bytes) {
+		*out = NULL;
+		return 0;
+	}
+	if (offset + bytes > view->bytes)
+		bytes = view->bytes - offset;
+	return text_raw(view->text, out, view->start + offset, bytes);
+}
+
+size_t view_delete(struct view *view, position_t offset, size_t bytes)
 {
 	return text_delete(view->text, view->start + offset, bytes);
 }
 
-unsigned view_insert(struct view *view, const void *in,
-		     unsigned offset, int bytes)
+size_t view_insert(struct view *view, const void *in,
+		   position_t offset, ssize_t bytes)
 {
 	if (bytes < 0)
 		bytes = in ? strlen(in) : 0;
